#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2023 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

from argparse import RawTextHelpFormatter
import argparse
import logging
import sys
import traceback

from flame.license_db import FossLicenses
from flame.license_db import Validation
import flame.config
from flame.format import OUTPUT_FORMATS
from flame.format import OutputFormatterFactory
from flame.exception import FlameException

def get_parser():

    parser = argparse.ArgumentParser(
        description=flame.config.DESCRIPTION,
        epilog=flame.config.EPILOG,
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument('-fc', '--flame-config',
                        type=str,
                        dest='flame_config',
                        help='Read settings from configuration file.',
                        default=None)

    parser.add_argument('-ld', '--license-dir',
                        type=str,
                        dest='license_dir',
                        help='Instead of using flame\'s built in licenses, use the licenses in the supplied directory',
                        default=None)

    parser.add_argument('-ald', '--additional-license-dir',
                        type=str,
                        dest='additional_license_dir',
                        help='Add licenses as found in the supplied directory',
                        default=None)

    parser.add_argument('-lmf', '--license-matrix-file',
                        type=str,
                        dest='license_matrix_file',
                        help='Supply license matrix to override OSADL\'s license compatibility matrix',
                        default=None)

    parser.add_argument('-ndu', '--no-dual-update',
                        action='store_true',
                        help='do not update dual licenses (e.g. GPL-2.0-or-later -> GPL-2.0-only OR GPL-3.0-only)',
                        default=False)

    parser.add_argument('-of', '--output-format',
                        type=str,
                        help=f'Chose output format. Available formats: {", ".join(OUTPUT_FORMATS)}',
                        default='text')

    parser.add_argument('-v', '--verbose',
                        action='store_true',
                        help='output verbose information',
                        default=False)

    parser.add_argument('-V', '--version',
                        action='store_true',
                        help='output version information',
                        default=False)

    parser.set_defaults(which='help', func=version_info)

    parser.add_argument('-d', '--debug',
                        action='store_true',
                        help='output debug information to stderr',
                        default=False)

    parser.add_argument('-c', '--check',
                        action='store_true',
                        help='check all license files against JSON schema when initializing. Mainly for the flame developers.',
                        default=False)

    subparsers = parser.add_subparsers(help='Sub commands')
    parser.add_argument('--validate-spdx', action='store_true', dest='validate_spdx', help='Validate that the resulting license expression is valid according to SPDX syntax', default=False)
    parser.add_argument('--validate-scancode', action='store_true', dest='validate_scancode', help='Validate that the resulting license expression is valid according to SPDX syntax and identifier exists in the Scancode license list', default=False)
    parser.add_argument('--validate-relaxed', action='store_true', dest='validate_relaxed', help='Validate that the resulting license expression is valid according to SPDX syntax, but allow non SPDX identifiers ', default=False)
    parser.add_argument('--validate-osadl', action='store_true', dest='validate_osadl', help='Validate that the resulting licenses are supported by OSADL\'s compatibility matrix.', default=False)

    # license
    parser_e = subparsers.add_parser(
        'license', help='Convert license to SPDX identifiers and syntax')
    parser_e.set_defaults(which='license', func=show_license)
    parser_e.add_argument('license', type=str, nargs='+', help='license expression to fix')

    # full license
    parser_e = subparsers.add_parser(
        'license-full', help='Display full information about the license')
    parser_e.set_defaults(which='license-full', func=full_license)
    parser_e.add_argument('license', type=str, help='license to display fully. Only single license is allowed')

    # compatibility
    parser_c = subparsers.add_parser(
        'compat', help='Convert license to using licenses existing in the OSADL\'s matrix')
    parser_c.set_defaults(which='compat', func=compatibility)
    parser_c.add_argument('license', type=str, nargs='+', help='license name to display')

    # simplify
    parser_s = subparsers.add_parser(
        'simplify', help='Simplify a license (SPDX syntax) expression, e.g. "MIT AND MIT" -> "MIT"')
    parser_s.set_defaults(which='simplify', func=simplify)
    parser_s.add_argument('license_expression', type=str, nargs='+', help='license expression to simplify')

    # aliases
    parser_a = subparsers.add_parser(
        'aliases', help='Display all aliases')
    parser_a.set_defaults(which='aliases', func=aliases)
    parser_a.add_argument('--include-license', '-il', type=str, dest='include_license', help='List only the aliases for licenses containing the given string (case insensitive)', default=None)

    # ambiguities
    parser_os = subparsers.add_parser(
        'ambiguities', help='Display all ambiguities')
    parser_os.set_defaults(which='ambiguities', func=ambiguities)

    # compatbilities
    parser_cs = subparsers.add_parser(
        'compats', help='Display all compatibilities')
    parser_cs.set_defaults(which='compats', func=compats)

    # compounds
    parser_cs = subparsers.add_parser(
        'compounds', help='Display all compounds')
    parser_cs.set_defaults(which='compounds', func=compounds)

    # licenses
    parser_cs = subparsers.add_parser(
        'licenses', help='show all licenses')
    parser_cs.set_defaults(which='licenses', func=licenses)

    # operators
    parser_os = subparsers.add_parser(
        'operators', help='Display all operators')
    parser_os.set_defaults(which='operators', func=operators)

    # unknown
    parser_u = subparsers.add_parser(
        'unknown', help='Show the unknown licenses for a license expression. Intended for foss-licenses developers.')
    parser_u.set_defaults(which='unknown', func=unknown)
    parser_u.add_argument('license', type=str, nargs='+', help='license expression to fix')

    return parser

def parse():

    return get_parser().parse_args()

def ambiguities(fl, formatter, args):
    all_ambiguities = fl.ambiguities_list()
    return formatter.format_ambiguities(all_ambiguities)

def operators(fl, formatter, args):
    all_op = fl.operators()
    return formatter.format_operators(all_op, args.verbose)

def aliases(fl, formatter, args):
    all_aliases = fl.alias_list(args.include_license)
    return formatter.format_alias_list(all_aliases, args.verbose)

def licenses(fl, formatter, args):
    all_licenses = fl.licenses()
    return formatter.format_licenses(all_licenses, args.verbose)

def unknown(fl, formatter, args):
    validations = __validations(args)
    return fl.unknown_symbols(args.license, validations)

def simplify(fl, formatter, args):
    validations = __validations(args)
    expression = fl.expression_license(' '.join(args.license_expression), validations=validations, update_dual=(not args.no_dual_update))
    simplified = fl.simplify([expression['identified_license']])
    return formatter.format_licenses([str(simplified)], args.verbose)

def compats(fl, formatter, args):
    all_compats = fl.compatibility_as_list()
    return formatter.format_compat_list(all_compats, args.verbose)

def compounds(fl, formatter, args):
    compounds = fl.compound_list()
    return formatter.format_compounds(compounds, args.verbose)

def version_info(fl, formatter, args):
    return flame.config.SW_VERSION, None

def compatibility(fl, formatter, args):
    validations = __validations(args)
    compatibilities = fl.expression_compatibility_as(' '.join(args.license), validations, update_dual=(not args.no_dual_update))
    return formatter.format_compatibilities(compatibilities, args.verbose)

def show_license(fl, formatter, args):
    validations = __validations(args)
    expression = fl.expression_license(' '.join(args.license), validations=validations, update_dual=(not args.no_dual_update))
    return formatter.format_expression(expression, args.verbose)

def full_license(fl, formatter, args):
    expr_split = fl.expression_license(args.license)['identified_license'].split()

    if len(expr_split) == 1:
        lic = fl.license_complete(expr_split[0])
        return formatter.format_license_complete(lic, args.verbose)
    else:
        raise FlameException(f'You can only provide one license to license-full. "{args.license} not allowed"')

def __validations(args):
    validations = []
    if args.validate_spdx:
        validations.append(Validation.SPDX)
    if args.validate_relaxed:
        validations.append(Validation.RELAXED)
    if args.validate_osadl:
        validations.append(Validation.OSADL)
    if args.validate_scancode:
        validations.append(Validation.SCANCODE)

    return validations

def main():

    args = parse()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)

    formatter = OutputFormatterFactory.formatter(args.output_format)

    config = {}
    config['flame-config'] = args.flame_config
    config['check'] = args.check
    config['additional-license-dir'] = args.additional_license_dir
    config['license-dir'] = args.license_dir
    config['flame-config'] = args.flame_config
    config['license-matrix-file'] = args.license_matrix_file

    try:
        fl = FossLicenses(config=config)
    except Exception as e:
        if args.verbose:
            print(traceback.format_exc())
            print(str(e))

        formatted = formatter.format_error(e, args.verbose)
        print(formatted)
        sys.exit(1)

    if args.func:
        try:
            formatted, warnings = args.func(fl, formatter, args)
            if warnings:
                print(warnings, file=sys.stderr)
            print(formatted)
        except Exception as e:
            if args.verbose:
                print(traceback.format_exc(), file=sys.stderr)

            formatted, warnings = formatter.format_error(e, args.verbose)
            print(formatted)
            sys.exit(1)


if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2024 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import boolean
import collections
import glob
import json
import logging
import re
from pathlib import Path
import license_expression
import spdx_license_list
from enum import Enum

from flame.config import LICENSE_DIR, LICENSE_SCHEMA_FILE, read_config
from flame.config import LICENSE_OPERATORS_FILE, LICENSE_COMPOUNDS_FILE, LICENSE_AMBIG_FILE, LICENSE_DUALS_FILE

from flame.exception import FlameException
from jsonschema import validate

import osadl_matrix

json_schema = None

FLAME_ALIASES_TAG = 'aliases'
FLAME_COMPATIBLE_LICENSE_TAG = 'compat_license'
FLAME_IDENTIFIED_LICENSE_TAG = 'identified_license'
FLAME_LICENSE_TEXT_TAG = 'license_text'
FLAME_NAME_TAG = 'name'
FLAME_COMPATIBILITY_TAG = 'compatibility'

COMPATIBILITY_AS_TAG = 'compatibility_as'
IDENTIFIED_ELEMENT_TAG = 'identified_element'
SCANCODE_KEY_TAG = 'scancode_key'
SCANCODE_KEYS_TAG = 'scancode_keys'
LICENSES_TAG = 'licenses'
COMPATS_TAG = 'compats'
DUALS_TAG = 'dual-licenses'
AMBIG_TAG = 'ambiguties'
DUAL_LICENSES_TAG = 'dual-licenses'
DUAL_NEWER_TAG = 'newer-versions'
COMPOUNDS_TAG = 'compounds'

LICENSE_OPERATORS_TAG = 'license_operators'

# expression for splitting a license expression in to a list of the
# actual license name/id without the "AND", "OR", "(", ")"
#
# Note: OR and AND needs a space around them to prevent splitting for
# example "OReilly" in to "OR eilly"
LICENSE_SPLIT_RE_CLEAN = r' AND | OR |\(|\)'
LICENSE_SPLIT_RE = r'( AND | OR |\(|\))'

class Validation(Enum):
    RELAXED = 1
    SPDX = 2
    OSADL = 3
    SCANCODE = 4

class FossLicenses:
    """
    Return a FossLicenses object.
    The config object is checked for the follow variables:
    * check (boolean): enable check of each license against schema
    * license-dir (str): directory where licenses (JSON and LICENSE) are located. Used for testing.
    * additional-license-dir (str): add directory to licenses (JSON and LICENSE) are located. Used for extending flame.
    * logging-level (str): log level to use
    * flame-config (str): configuration file to read settings from


    :param config:
    :raise FlameException: if license_expression is not valid
    :Example:

    >>> fl = FossLicenses()

    """
    def __init__(self, config=None):
        if not config:
            config = {}
        # get config from passed file name via config parameter
        config_from_file = read_config(config.get('flame-config', None))
        # read config from file passed via config parameter
        config_from_file.update((k, v) for k, v in config.items() if v is not None)
        config = config_from_file

        check = config.get('check', False)
        logging_level = self.__str_to_loggin_info(config)
        logging.basicConfig(level=logging_level)

        self.config = config
        self.license_dir = config.get('license-dir', LICENSE_DIR)
        self.license_matrix_file = config.get('license-matrix-file', None)
        self.additional_license_dir = config.get('additional-license-dir', [])
        self.__init_license_db(check)
        self.supported_licenses = None
        self.compat_cache = {}
        self.license_cache = {}

    def __str_to_loggin_info(self, config):
        logging_level = config.get('level')
        if not logging_level:
            return logging.INFO
        if logging_level.lower() == "info":
            return logging.INFO
        elif logging_level.lower() == "warning":
            return logging.WARNING
        elif logging_level.lower() == "debug":
            return logging.DEBUG

    def __read_json(self, file_name):
        with open(file_name) as f:
            return json.load(f)

    def __validate_license_data(self, license_data):
        global json_schema
        if not json_schema:
            schema_file = LICENSE_SCHEMA_FILE
            logging.debug(f'Reading JSON schema from {schema_file}')
            json_schema = self.__read_json(schema_file)
        validate(instance=license_data, schema=json_schema)

    def __read_license_file(self, license_file, check=False):
        data = self.__read_json(license_file)
        if check:
            self.__validate_license_data(data)

        license_text_file = license_file.replace('.json', '.LICENSE')
        if not Path(license_text_file).is_file():
            raise FileNotFoundError(f'Could not find "{license_text_file}" matching "{license_file}"')
        with open(license_text_file) as lf:
            data[FLAME_LICENSE_TEXT_TAG] = lf.read()

        return data

    def __init_license_db(self, check=False):
        self.license_db = {}
        licenses = {}
        aliases = {}
        duals = {}
        scancode_keys = {}
        compats = {}
        logging.debug(f'reading from: {self.license_dir}')
        license_dirs = [self.license_dir]
        self.ambiguities = {'ambiguities': [], 'aliases': {}}

        if self.additional_license_dir:
            license_dirs.append(self.additional_license_dir)
        for license_dir in license_dirs:
            for license_file in glob.glob(f'{license_dir}/*.json'):
                if "duals" in license_file:
                    continue
                if "compounds" in license_file:
                    continue
                if "ambig" in license_file:
                    continue
                logging.debug(f'license_file: {license_file}')
                data = self.__read_license_file(license_file, check)
                licenses[data['spdxid']] = data
                for alias in data[FLAME_ALIASES_TAG]:
                    if alias in aliases:
                        raise FlameException(f'Alias "{alias}" -> {data["spdxid"]} already defined as "{aliases[alias]}".')

                    aliases[alias] = data['spdxid']
                if SCANCODE_KEY_TAG in data:
                    scancode_keys[data[SCANCODE_KEY_TAG]] = data['spdxid']
                if COMPATIBILITY_AS_TAG in data:
                    compats[data['spdxid']] = data[COMPATIBILITY_AS_TAG]

        # Ambiguous licenses
        ambig_file = self.config.get('ambiguity_file', LICENSE_AMBIG_FILE)
        logging.debug(f' * ambiguities file: {ambig_file}')
        data = self.__read_json(ambig_file)
        data['aliases'] = {}

        for k, v in data['ambiguities'].items():
            # for quicker lookups, add 'aliases' which is the reverse of
            # the aliases per license I.e a quicker lookup table when
            # identifying ambiguous licenses
            for alias in v['aliases']:
                data['aliases'][alias] = k
            data['aliases'][k] = k
        self.ambiguities = data

        # Compound licenses
        # some compound licenses are incorrectly stated as
        # one, e.g. "GPL-2.0-with-classpath-exception" which
        # should be "GPL-2.0-only WITH
        # Classpath-exception-2.0". This file provides
        # translations for such
        compounds_file = self.config.get('compounds_file', LICENSE_COMPOUNDS_FILE)
        data = self.__read_json(compounds_file)
        self.license_db[COMPOUNDS_TAG] = data['compounds']
        for compound in data[COMPOUNDS_TAG]:
            licenses[compound['spdxid']] = compound
            for alias in compound[FLAME_ALIASES_TAG]:
                if alias in aliases:
                    raise FlameException(f'Alias "{alias}" -> {compound["spdxid"]} already defined as "{aliases[alias]}".')

                aliases[alias] = compound['spdxid']

            if COMPATIBILITY_AS_TAG in compound:
                compats[compound['spdxid']] = compound[COMPATIBILITY_AS_TAG]

        # Dual licenses
        # Read license with built-in dual feature, e.g
        # "GPL-2.0-or-later" which can be seen as a dual
        # license "GPL-2.0-only OR GPL-3.0-only"
        duals_file = self.config.get('duals_file', LICENSE_DUALS_FILE)
        data = self.__read_json(duals_file)
        for dual in data[DUAL_LICENSES_TAG]:
            duals[dual['spdxid']] = dual

        logging.debug(f'compounds_file: {compounds_file}')
        logging.debug(f'ambig_file: {ambig_file}')
        logging.debug(f'duals_file: {duals_file}')
        logging.debug(f'config: {self.config}')

        self.license_expression = license_expression.get_spdx_licensing()
        self.license_db[DUALS_TAG] = duals
        self.license_db[AMBIG_TAG] = self.ambiguities
        self.license_db[LICENSES_TAG] = licenses
        self.license_db[COMPATS_TAG] = compats
        self.license_db[FLAME_ALIASES_TAG] = aliases
        self.license_db[SCANCODE_KEYS_TAG] = scancode_keys
        self.license_db[LICENSE_OPERATORS_TAG] = self.__read_json(LICENSE_OPERATORS_FILE)['operators']

    def __identify_license(self, name):
        if name in self.license_db[LICENSES_TAG]:
            ret_name = name
            ret_id = 'direct'
        elif name in self.license_db[LICENSE_OPERATORS_TAG]:
            ret_name = self.license_db[LICENSE_OPERATORS_TAG][name]
            ret_id = 'operator'
        elif name in self.license_db[FLAME_ALIASES_TAG]:
            ret_name = self.license_db[FLAME_ALIASES_TAG][name]
            ret_id = 'alias'
        else:
            raise FlameException(f'Could not identify license from "{name}"')

        return {
            'queried_name': name,
            'name': ret_name,
            'identified_via': ret_id,
        }

    def __update_license_expression_helper(self, needles, needle_tag, license_expression, allow_letter=False):
        replacements = []
        for needle in reversed(collections.OrderedDict(sorted(needles.items(), key=lambda x: len(x[0])))):
            if allow_letter:
                reg_exp = r'( |\(|^|\)|\||/)%s( |$|\)|\||&|[a-zA-Z])' % re.escape(needle)
                extra_add = ' '
            else:
                reg_exp = r'( |\(|^|\)|\||/)%s( |$|\)|\||&)' % re.escape(needle)
                extra_add = ''
            if re.search(reg_exp, license_expression):
                replacement = needles[needle]
                replacements.append({
                    'queried_name': needle,
                    'name': replacement,
                    'identified_via': needle_tag,
                })
                license_expression = re.sub(reg_exp, f'\\1 {replacement}{extra_add}\\2', license_expression)
        return {
            'license_expression': re.sub(r'\s\s*', ' ', license_expression).strip(),
            'identifications': replacements,
        }

    def __update_or_later(self, license_expression):
        updates = []
        new_expr = []
        for lic in re.split(LICENSE_SPLIT_RE, license_expression):
            if lic.strip() == '':
                continue
            trimmed_license = lic.strip().split()[0]
            if trimmed_license in self.license_db[DUALS_TAG]:
                newer_list = self.license_db[DUALS_TAG][trimmed_license][DUAL_NEWER_TAG]
                inner_new = []
                for newer in newer_list:
                    if len(inner_new) > 0:
                        inner_new.append('OR')
                    inner_new.append(lic.replace(trimmed_license, newer))
                inner_new_str = f' ( {" ".join(inner_new)} ) '
                updates.append({
                    'license-expression': lic,
                    'license': trimmed_license,
                    'updates': newer_list,
                    'updated-license': inner_new_str,
                })

                new_expr.append(inner_new_str)
            else:
                new_expr.append(lic)
        ret = ' '.join(new_expr)
        return {
            'input_license_expression': license_expression,
            'license_expression': ret,
            'updates': updates,
        }

    def expression_license(self, license_expression, validations=None, update_dual=True):
        """
        Return an object with information about the normalized license for the license given.

        :param license_expression: A license expression. E.g "BSD3" or "GPLv2+ || BSD3"
        :type license_expression: str
        :raise FlameException: if license_expression is not valid
        :return: a normalized license expression
        :rtype: list

        :Example:

        >>> fl = FossLicenses()
        >>> expression = fl.expression_license('BSD3 & x11-keith-packard')
        >>> print(expression['identified_license'])
        BSD-3-Clause AND LicenseRef-flame-x11-keith-packard

        """
        if not isinstance(license_expression, str):
            raise FlameException('Wrong type (type(license_expresssion)) of input to the function expression_license. Only string is allowed.')

        # remove multiple blanks
        license_expression = re.sub(' [ ]*', ' ', license_expression)

        cache_key = f'{license_expression}__{validations}__{update_dual}'
        if cache_key in self.license_cache:
            return self.license_cache.get(cache_key)

        replacements = []

        ret = self.__update_license_expression_helper(self.license_db[FLAME_ALIASES_TAG],
                                                      'alias',
                                                      license_expression)
        replacements += ret['identifications']
        ret = self.__update_license_expression_helper(self.license_db[LICENSE_OPERATORS_TAG],
                                                      'operator',
                                                      ret['license_expression'],
                                                      allow_letter=True)
        replacements += ret['identifications']

        # manage ambiguities
        ambiguities = []
        aliases = self.license_db[AMBIG_TAG]['aliases']

        tmp_license_expression = ret['license_expression']
        for alias in reversed(collections.OrderedDict(sorted(aliases.items(), key=lambda x: len(x[0])))):
            needle = r'(?:\s+|^)%s(?:\s+|$)' % re.escape(alias.replace(" and ", " AND "))
            if re.search(needle, tmp_license_expression):
                real_lic = self.license_db[AMBIG_TAG]['aliases'][alias]
                if alias != real_lic:
                    about_license = f'An ambiguity was identified in "{ret["license_expression"]}". The ambiguous license is "{real_lic}", identified via "{alias}".'
                else:
                    about_license = f'An ambiguity was identified in "{ret["license_expression"]}". The ambiguous license is "{real_lic}"-'
                problem = self.license_db[AMBIG_TAG]["ambiguities"][real_lic]["problem"]
                ambiguities.append({
                    'license': ret['license_expression'],
                    'ambigous_license': real_lic,
                    'problem': problem,
                    'description': f'{about_license} Problem: {problem}',
                })
                tmp_license_expression = re.sub(needle, ' ', tmp_license_expression)

        update_problem = None
        updates = []
        if update_dual:
            # Manage dual licenses (such as GPL-2.0-or-later)
            updates_object = self.__update_or_later(ret['license_expression'])
            updates = updates_object['updates']
            try:
                updated_license = str(self.license_expression.parse(updates_object['license_expression']))
            except boolean.boolean.ParseError as e:
                update_problem = f'Could not parse \"{updates_object["license_expression"]}\". Exception: {e}'
                updated_license = updates_object['license_expression']
        else:
            try:
                updated_license = str(self.license_expression.parse(ret['license_expression']))
            except boolean.boolean.ParseError as e:
                update_problem = f'Could not parse \"{updates_object["license_expression"]}\". Exception: {e}'
                updated_license = updates_object['license_expression']

        license_parsed = updated_license

        self.__validate_license(validations, license_parsed)

        if update_problem:
            amb_text = ''
            if ambiguities:
                amb_text = f'Ambiguities were found: {ambiguities}.'

            parse_text = ''
            if update_problem:
                parse_text = f'Parsing failed: {update_problem}.'

            raise FlameException(f'Could not parse the license "{license_expression}". {amb_text} {parse_text}')

        ret = {
            'queried_license': license_expression,
            FLAME_IDENTIFIED_LICENSE_TAG: license_parsed,
            'identifications': replacements,
            'ambiguities': ambiguities,
            'updated_license': updated_license,
            'license_parsed': license_parsed,
            'updates': updates,
        }
        self.license_cache[cache_key] = ret
        return ret

    def licenses(self):
        """
        Returns all licenses supported by flame

        :Example:

        >>> fl = FossLicenses()
        >>> expression = fl.licenses()

        """
        return list(self.license_db[LICENSES_TAG].keys())

    def license_complete(self, name):
        """
        Return the corresponding license object

        :param name: spdx identifier, alias or scancode key2
        :type name: str
        :raise FlameException: if license_expression is not valid

        :Example:

        >>> fl = FossLicenses()
        >>> expression = fl.license_complete("MIT")

        """
        identified_name = self.__identify_license(name)['name']
        return self.license_db[LICENSES_TAG][identified_name]

    def license(self, name):
        """
        Return the normalized license name (SPDXID)

        :param name: spdx identifier, alias or scancode key2
        :type name: str
        :raise FlameException: if license_expression is not valid
        :return: SPDX identifier for the license 'name'
        :rtype: str

        :Example:

        >>> fl = FossLicenses()
        >>> license = fl.license('BSD3')
        >>> print(license['identified_element']['name'])
        BSD-3-Clause

        """

        identified_license = self.__identify_license(name)
        identified_name = identified_license[FLAME_NAME_TAG]
        if identified_license['identified_via'] == 'operator':
            return {
                IDENTIFIED_ELEMENT_TAG: identified_license,
                'operator': True,
            }
        else:
            return {
                IDENTIFIED_ELEMENT_TAG: identified_license,
                'license': self.license_db[LICENSES_TAG][identified_name],
            }

    def __OBSOLETE__license_spdxid(self, name):
        """
        name: spdx identifier, alias or scancode key2

        returns the corresponding spdxid
        """
        return self.license(name)['license']['spdxid']

    def __OBSOLETE__license_scancode_key(self, name):
        """
        name: spdx identifier, alias or scancode key2

        returns the corresponding scancode_key
        """
        return self.license(name)['license']['scancode_key']

    def compatibility_as_list(self) -> [str]:
        """Return a list of all the licenses missing in the OSADL matrix but having a known similar compatibility.

        :Example:

        >>> fl = FossLicenses()
        >>> compats = fl.compatibility_as_list()

        """
        # List all compatibility_as that exist
        licenses = self.license_db[LICENSES_TAG]
        return [{COMPATIBILITY_AS_TAG: licenses[x][COMPATIBILITY_AS_TAG], 'spdxid': licenses[x]['spdxid']} for x in licenses if COMPATIBILITY_AS_TAG in licenses[x]]

    def compound_list(self) -> [str]:
        """Return a list of all the compound licenses.

        :Example:

        >>> fl = FossLicenses()
        >>> compats = fl.compound_list()

        """
        # List all compounds
        return self.license_db[COMPOUNDS_TAG]

    def alias_list(self, alias_license: str = None) -> [str]:
        """Returns a list of all the aliases. Supplying alias_license
        will return a list of aliases beginning with alias_license

        :param str alias_license:  limit the list of alias to all matching alias_license

        :Example:

        >>> fl = FossLicenses()
        >>> aliases = fl.alias_list()

        """
        if alias_license:
            return {k: v for k, v in self.license_db[FLAME_ALIASES_TAG].items() if alias_license.lower() in v.lower()}
        # List all aliases that exist
        return self.license_db[FLAME_ALIASES_TAG]

    def ambiguities_list(self):
        """Returns a list of all the ambigious licenses.

        :Example:

        >>> fl = FossLicenses()
        >>> aliases = fl.ambiguities_list()

        """
        # List all aliases that exist
        return self.license_db[AMBIG_TAG]['ambiguities']

    def unknown_symbols(self, _license_expression, validations=None):
        """Returns a list of all all unknown license symbols for a license expression
        Note: developer function - not for "normal" users

        :Example:

        >>> fl = FossLicenses()
        >>> unknowns = fl.unknown_symbols(["MIT"])

        """
        license_expression = ' '.join(_license_expression)

        try:
            compat_license_expression = self.expression_license(license_expression, validations=validations, update_dual=False)
            fixed_license_expression = compat_license_expression['identified_license']
        except Exception as e:
            fixed_license_expression = license_expression
            ambig_aliases = {}
            for ambig in self.ambiguities_list():
                aliases = self.license_db[AMBIG_TAG]['ambiguities'][ambig]['aliases']
                for alias in aliases:
                    ambig_aliases[alias] = ambig

            for alias in reversed(collections.OrderedDict(sorted(ambig_aliases.items(), key=lambda x: len(x[0])))):
                if alias in fixed_license_expression:
                    fixed_license_expression = re.sub(re.escape(alias), ambig_aliases[alias], fixed_license_expression)
                    break
        compat_licenses = [x.strip() for x in re.split(LICENSE_SPLIT_RE_CLEAN, fixed_license_expression)]
        compat_licenses = [x for x in compat_licenses if x]
        unknown_symbols = set()
        for compat_license in compat_licenses:
            if compat_license not in self.known_symbols():
                unknown_symbols.add(compat_license)
        if len(unknown_symbols) != 0:
            raise FlameException('Unknown symbols identified.\n' + '\n'.join(list(unknown_symbols)))
        return 'OK', None

    def known_symbols(self):
        """Returns a list of all all known license symbols.

        :Example:

        >>> fl = FossLicenses()
        >>> symbols = fl.known_symbols()

        """
        _symbols = set()

        ambiguities = self.ambiguities_list()
        for ambig in ambiguities:
            _symbols.add(ambig)
            _symbols.update(set(ambiguities[ambig]['aliases']))

        licenses = self.license_db[LICENSES_TAG]
        for lic in licenses:
            _symbols.add(lic)
            _symbols.update(set(licenses[lic]['aliases']))

        operators = self.license_db[LICENSE_OPERATORS_TAG]
        for op in operators:
            _symbols.add(op)

        return list(_symbols)

    def aliases(self, license_name):
        """Returns a list of all the aliases for a license

        :param str license_name: Exact name (SPDXID) of the license

        :Example:

        >>> fl = FossLicenses()
        >>> aliases = fl.aliases("GPLv2+")

        """
        identified_name = self.__identify_license(license_name)[FLAME_NAME_TAG]
        return self.license_db[LICENSES_TAG][identified_name][FLAME_ALIASES_TAG]

    def operators(self):
        """
        Returns a list of all the supported (boolean) operators in license expressions.

        :Example:

        >>> fl = FossLicenses()
        >>> operators = fl.operators()

        """
        return self.license_db[LICENSE_OPERATORS_TAG]

    def __compatibility_as(self, license_name):
        # List compatibility_as for license
        identified = self.__identify_license(license_name)
        identified_name = identified[FLAME_NAME_TAG]

        if COMPATIBILITY_AS_TAG in self.license_db[LICENSES_TAG][identified_name]:
            compat = self.license_db[LICENSES_TAG][identified_name][COMPATIBILITY_AS_TAG]
            method = COMPATIBILITY_AS_TAG
        else:
            compat = identified_name
            method = 'direct'

        return {
            IDENTIFIED_ELEMENT_TAG: identified,
            FLAME_COMPATIBILITY_TAG: {
                'compat_as': compat,
                'queried_name': license_name,
                'identified_via': method,
            },
        }

    def simplify(self, expression):
        return self.license_expression.parse(' '.join(expression)).simplify()

    def expression_compatibility_as(self, license_expression, validations=None, update_dual=True):
        """Returns an object with information about the compatibility status for the license given.

        :param str license_expression: A license expression. E.g "BSD3" or "GPLv2+ || BSD3"

        :Example: supplying only one license, so look at [0]

        >>> fl = FossLicenses()
        >>> compat = fl.expression_compatibility_as('x11-keith-packard')
        >>> print(compat['compatibilities'][0]['name'])
        HPND

        """

        cache_key = f'{license_expression}__{validations}__{update_dual}'
        if cache_key in self.compat_cache:
            return self.compat_cache.get(cache_key)

        expression_full = self.expression_license(license_expression, validations, update_dual)
        compats = []
        ret = self.__update_license_expression_helper(self.license_db[COMPATS_TAG],
                                                      'compat',
                                                      expression_full[FLAME_IDENTIFIED_LICENSE_TAG])
        ret['license_expression'] = re.sub(r'\s\s*', ' ', ret['license_expression']).strip()
        compats = ret['identifications']
        compat_license_expression = ret['license_expression']

        compat_licenses = [x.strip() for x in re.split(LICENSE_SPLIT_RE, compat_license_expression)]
        compat_licenses = [x for x in compat_licenses if x]
        compat_support = self.__validate_compatibilities_support(compat_licenses)

        self.__validate_license(validations, compat_license_expression)

        ret = {
            'ambiguities': expression_full['ambiguities'],
            'compatibilities': compats,
            'queried_license': license_expression,
            'identification': expression_full,
            FLAME_IDENTIFIED_LICENSE_TAG: expression_full[FLAME_IDENTIFIED_LICENSE_TAG],
            FLAME_COMPATIBLE_LICENSE_TAG: compat_license_expression,
            'compat_support': compat_support,
        }
        self.compat_cache[cache_key] = ret
        return ret

    def __validate_license(self, validations, license_expression):
        if validations:
            if Validation.SPDX in validations:
                self.__validate_license_spdx(license_expression)
            if Validation.SCANCODE in validations:
                self.__validate_license_scancode(license_expression)
            if Validation.RELAXED in validations:
                self.__validate_license_relaxed(license_expression)
            if Validation.OSADL in validations:
                compat_license_expression = self.expression_compatibility_as(license_expression)['compat_license']
                compat_licenses = [x.strip() for x in re.split(LICENSE_SPLIT_RE_CLEAN, compat_license_expression)]
                compat_licenses = [x for x in compat_licenses if x]
                compat_support = self.__validate_compatibilities_support(compat_licenses)
                self.__validate_licenses_osadl(compat_support)

    def __validate_compatibilities_support(self, licenses):
        compat_support = {}
        compat_support['licenses'] = []
        all_supported = True
        for lic in licenses:
            support = self.__validate_compatibility_support(lic)
            compat_support['licenses'].append({
                'license': lic,
                'supported': support,
            })
            all_supported = all_supported and support
        compat_support['supported'] = all_supported

        return compat_support

    def __validate_compatibility_support(self, lic):
        if not self.supported_licenses:
            self.support_licenses = osadl_matrix.supported_licenses(self.license_matrix_file)
        return lic in self.support_licenses

    def __validate_license_spdx(self, expr):
        """
        """
        missing = []
        license_list = self.__license_list(expr)
        for _lic in license_list:
            lic = _lic.strip()
            spdx_license = lic in spdx_license_list.LICENSES
            spdx_exception = "exception" in lic.lower()
            if not (spdx_license or spdx_exception):
                missing.append(lic)

        if missing:
            raise FlameException(f'License validation of "{expr}" failed. Non SPDX identifiers: "{", ".join(missing)}"')

    def __validate_license_scancode(self, expr):
        """
        """
        expr_info = self.license_expression.validate(expr)

        if expr_info.errors:
            raise FlameException(f'License validation of "{expr}" failed. Errors: "{", ".join(expr_info.errors)}"')

    def __validate_license_relaxed(self, expr):
        """
        """
        license_list = self.__license_list(expr)
        for _lic in license_list:
            lic = _lic.strip()
            if " " in lic.strip():
                raise FlameException(f'Found license with multiple words "{lic}"')

    def __validate_licenses_osadl(self, compat_supported):
        """
        """
        if not compat_supported['supported']:
            raise FlameException('Not all licenses supported by OSADL\'s compatibility matrix',
                                 compat_supported)

    def __license_list(self, expr):
        SPDX_OPERATORS = ['AND', 'OR', 'WITH']
        return re.split(f'{"|".join(SPDX_OPERATORS)}', expr)



# SPDX-FileCopyrightText: 2023 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

class FlameException(Exception):
    """
    Simple exception for module
    """
    def __init__(self, message, problems=None):
        super().__init__(message)
        self._problems = problems

    def problems(self):
        if self._problems:
            return self._problems
        else:
            return ''



# SPDX-FileCopyrightText: 2023 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging
import os
import flame.exception

SW_VERSION = '0.20.10'

PYTHON_DIR = os.path.dirname(os.path.realpath(__file__))
VAR_DIR = os.path.join(PYTHON_DIR, 'var')
LICENSE_DIR = os.path.join(VAR_DIR, 'licenses')

LICENSE_SCHEMA_FILE = os.path.join(VAR_DIR, 'license_schema.json')
LICENSE_OPERATORS_FILE = os.path.join(VAR_DIR, 'operators.json')
LICENSE_COMPOUNDS_FILE = os.path.join(VAR_DIR, 'compounds.json')
LICENSE_AMBIG_FILE = os.path.join(VAR_DIR, 'ambiguities.json')
LICENSE_DUALS_FILE = os.path.join(VAR_DIR, 'duals.json')

DESCRIPTION = """
NAME
  flame (FOSS License Additional Metadata) is a python API and
  command line tool

DESCRIPTION
  FOSS Licenses is a database with additional metadata about, primarily, FOSS licenses.

"""

EPILOG = f"""
CONFIGURATION
  All config files can be found in
  {VAR_DIR}

AUTHOR
  Henrik Sandklef

PROJECT SITE
  https://github.com/hesa/flame

REPORTING BUGS
  File a ticket at https://github.com/hesa/flame/issues

COPYRIGHT
  Copyright (c) 2024 Henrik Sandklef<hesa@sandklef.com>.
  License GPL-3.0-or-later

ATTRIBUTION
"""

def __read_config_helper(path):
    logging.debug(f'try reading: {path}')
    try:
        with open(path) as fp:
            logging.debug('data found')
            return json.load(fp)
    except Exception as e:
        logging.debug(f'file not found or usable: {path}')
        logging.debug(f'exception: {e}')
    raise flame.exception.FlameException(f'Could not read file: {path}')

def read_config(config_file=None):
    logging.debug('reading config')
    for path in [
            config_file,
            os.environ.get('FLAME_USER_CONFIG', None),
    ]:
        logging.debug(f'reading {path}')
        if path:
            try:
                return __read_config_helper(path)
            except Exception as e:
                logging.debug(f'reading {path} failed with {e}')
                raise flame.exception.FlameException(f'Could not read user specified file: {path}')

    for path in [
            os.path.join(os.environ.get('HOME', '/does/not/exist'), '.floss-flame.cfg'),
    ]:
        logging.debug(f'reading {path}')
        try:
            return __read_config_helper(config_file)
        except Exception as e:
            logging.debug(f'reading {path} failed with {e}')
            pass
    return {}



# SPDX-FileCopyrightText: 2024 Henrik Sandklef
#
# SPDX-License-Identifier: GPL-3.0-or-later


import json
import yaml

OUTPUT_FORMAT_JSON = 'Json'
OUTPUT_FORMAT_YAML = 'Yaml'
OUTPUT_FORMAT_TEXT = 'text'
OUTPUT_FORMATS = [OUTPUT_FORMAT_JSON, OUTPUT_FORMAT_YAML, OUTPUT_FORMAT_TEXT]
"""Allowed formats (case insensitive): Json, Yaml, text."""

class OutputFormatterFactory():

    @staticmethod
    def formatter(_format):
        """
        Return a formatter corresponding to the supplied format

        :param format: format for the output
        :type format: str
        :return: a normalized license expression, None if invalied format
        :rtype: OutputFormatter

        :Example:

        >>> formatter = OutputFormatterFactory.formatter("JSON")
        """
        if _format.lower() == OUTPUT_FORMAT_JSON.lower():
            return JsonOutputFormatter()
        elif _format.lower() == OUTPUT_FORMAT_YAML.lower():
            return YamlOutputFormatter()
        elif _format.lower() == OUTPUT_FORMAT_TEXT.lower():
            return TextOutputFormatter()

class OutputFormatter():
    """Base class for implementing sub classes providing methods for
    formatting data as returned from the various methods in the
    FossLicense class.

    There are no constructors, please use  :func:`flame.format.OutputFormatterFactory.formatter` instead.  See OUTPUT_FORMATS
    """

    def format_compat(self, compat, verbose=False):
        """
        Return a formatted string of the compatibility as returned by :func:`flame.license_db.FossLicenses.expression_compatibility_as`

        :param compat: A compatibitility as returned by expression_compatibility_as.
        :type compat: dict
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if compat is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> from flame.format import OutputFormatterFactory
        >>> fl = FossLicenses()
        >>> compat = fl.expression_compatibility_as('x11-keith-packard')
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted, warnings = formatter.format_compatibilities(compat)
        >>> print(formatted)
        HPND
        """
        return None, None

    def format_compat_list(self, all_compats, verbose=False):
        """
        Return a formatted string of the compatibilities :func:`flame.license_db.FossLicenses.compatibility_as_list`.

        :param all_compats: A list of compatibitility as returned from compatibility_as_list
        :type all_compats: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if all_compats is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> compat = fl.compatibility_as_list()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_compat_list(compat)

        """
        return None, None

    def format_expression(self, expression, verbose=False):
        """
        Return a formatted string of the compatibilities :func:`flame.license_db.FossLicenses.alias_list`.

        :param all_aliases: A list of aliases as returned from FossLicenses.alias_list()
        :type all_aliases: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if all_compats is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> compat = fl.alias_list()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_alias_list(compat)

        """
        return None, None

    def format_alias_list(self, all_aliases, verbose=False):
        """
        Return a formatted string of the compatibilities :func:`flame.license_db.FossLicenses.alias_list`.

        :param all_aliases: A list of aliases as returned from FossLicenses.alias_list()
        :type all_aliases: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if all_compats is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> compat = fl.alias_list()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_alias_list(compat)

        """
        return None, None

    def format_error(self, message, verbose=False):
        """
        Return a formatted string of an error message

        :param message: Error message, typically received via an exception
        :type message: str
        :param verbose: provide additional information
        :type verbose: boolean
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.exception import FlameException
        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> try:
        ...     aliases = fl.aliases("do no exist")
        ... except FlameException as e:
        ...     formatted = formatter.format_error(e)

        """
        return None, None

    def format_licenses(self, licenses, verbose=False):
        """
        Return a formatted string of a list of licenses

        :param licenses: list of licenses
        :type message: [str]
        :param verbose: provide additional information
        :type verbose: boolean
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> licenses = fl.licenses()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_licenses(licenses)

        """
        return None, None

    def format_license_complete(self, lic, verbose=False):
        """
        Return a formatted string of a complete license listing

        :param lic: license to format
        :type message: [str]
        :param verbose: provide additional information
        :type verbose: boolean
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> licenses = fl.license_complete("MIT")
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_license_complete(licenses)

        """
        return None, None

    def format_compatibilities(self, compats, verbose=False):
        """
        Return a formatted string of the provided compatibilities.

        :param : A list of aliases as returned from FossLicenses.alias_list()
        :type all_aliases: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if compats is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> compatibilities = fl.expression_compatibility_as("MIT & GPLv2+")
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_compatibilities(compatibilities)

        """
        return None, None

    def format_operators(self, operators, verbose=False):
        """
        Return a formatted string of the provided operators

        :param operators: A list of operators.
        :type operators: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if operators is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> operators = fl.operators()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_operators(operators)

        """
        return None, None

    def format_ambiguities(self, ambiguities, verbose=False):
        """
        Return a formatted string of the existing ambiguities

        :param ambiguities: A list of ambiguites.
        :type ambiguities: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if ambiguites is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> ambiguities = fl.ambiguities_list()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_ambiguities(ambiguities)

        """
        return None, None

    def format_compounds(self, compounds, verbose=False):
        """
        Return a formatted string of the existing compounds

        :param compounds: A list of ambiguites.
        :type compounds: list
        :param verbose: provide additional information
        :type verbose: boolean
        :raise FlameException: if compounds is not valid
        :return: formatted string
        :rtype: str

        :Example:

        >>> from flame.license_db import FossLicenses
        >>> fl = FossLicenses()
        >>> compounds = fl.compound_list()
        >>> formatter = OutputFormatterFactory.formatter("TEXT")
        >>> formatted = formatter.format_compounds(compounds)

        """
        return None, None

class JsonOutputFormatter(OutputFormatter):

    def format_compat(self, compat, verbose=False):
        return json.dumps(compat, indent=4), None

    def format_compat_list(self, all_compats, verbose=False):
        return json.dumps(all_compats, indent=4), None

    def format_expression(self, expression, verbose=False):
        return json.dumps(expression, indent=4), None

    def format_alias_list(self, all_aliases, verbose=False):
        return json.dumps(all_aliases, indent=4), None

    def format_error(self, error, verbose=False):
        new_error = {
            'message': str(error),
            'problems': error.problems(),
        }
        return json.dumps(new_error, indent=4), None

    def format_licenses(self, licenses, verbose=False):
        return json.dumps(licenses, indent=4), None

    def format_license_complete(self, _license, verbose=False):
        return json.dumps(_license, indent=4), None

    def format_compatibilities(self, compats, verbose=False):
        return json.dumps(compats), None

    def format_operators(self, operators, verbose=False):
        return json.dumps(operators), None

    def format_ambiguities(self, ambiguities, verbose=False):
        return json.dumps(ambiguities), None

    def format_compounds(self, compounds, verbose=False):
        return json.dumps(compounds), None

class YamlOutputFormatter(OutputFormatter):

    def format_compat(self, compat, verbose=False):
        return yaml.safe_dump(compat), None

    def format_compat_list(self, all_compats, verbose=False):
        return None, None

    def format_expression(self, expression, verbose=False):
        return yaml.safe_dump(expression), None

    def format_alias_list(self, all_aliases, verbose=False):
        return yaml.safe_dump(all_aliases), None

    def format_error(self, error, verbose=False):
        return yaml.safe_dump({'error': f'{error}'}), None

    def format_licenses(self, licenses, verbose=False):
        return yaml.safe_dump(licenses), None

    def format_license_complete(self, _license, verbose=False):
        return yaml.safe_dump(_license), None

    def format_compatibilities(self, compats, verbose=False):
        return yaml.safe_dump(compats), None

    def format_operators(self, operators, verbose=False):
        return yaml.safe_dump(operators), None

    def format_ambiguities(self, ambiguities, verbose=False):
        return yaml.safe_dump(ambiguities), None

    def format_compounds(self, compounds, verbose=False):
        return yaml.safe_dump(compounds), None

class TextOutputFormatter(OutputFormatter):

    def format_compat(self, compat, verbose=False):
        ret = []
        id_lic = compat['identified_license']
        if verbose:
            ret.append(f'queried_name: {id_lic["queried_name"]}')
            ret.append(f'name: {id_lic["name"]}')
            ret.append(f'identified via: {id_lic["identified_via"]}')
            ret.append(f'compatibility:  {compat["compatibility"]["compatibility"]}')
            ret.append(f'identified via: {compat["compatibility"]["identified_via"]}')
        else:
            ret.append(f'{compat["compatibility"]["compatibility"]}')
        return "\n".join(ret), None

    def format_compatibilities(self, compats, verbose=False):
        ret = []
        ret.append(f'{compats["compat_license"]}')
        if verbose:
            for compat in compats['compatibilities']:
                ret.append(f' * "{compat["queried_name"]}" -> "{compat["name"]}" via "{compat["identified_via"]}"')

        if compats['ambiguities']:
            descr = [x['description'] for x in compats['ambiguities']]
            warnings = f'Warnings: {", ".join(descr)}'
        else:
            warnings = None

        return '\n'.join(ret), warnings

    def format_compat_list(self, all_compats, verbose=False):
        return '\n'.join([f'{comp["spdxid"]} -> {comp["compatibility_as"]}' for comp in all_compats]), None

    def __OBSOLETE_format_identified(self, identified, verbose=False):
        ret = []
        id_lic = identified['identified_element']
        ret.append(f'{id_lic["name"]}')
        if verbose:
            ret.append(f' * "{id_lic["queried_name"]}" -> "{id_lic["name"]}" via "{id_lic["identified_via"]}"')
        return "\n".join(ret)

    def format_expression(self, expression, verbose=False):
        ret = []
        id_lic = expression['identified_license']
        ret.append(f'{id_lic}')
        if verbose:
            for identification in expression['identifications']:
                ret.append(f' * "{identification["queried_name"]}" -> "{identification["name"]}" via "{identification["identified_via"]}"')
        if expression['ambiguities']:
            descr = [x['description'] for x in expression['ambiguities']]
            warnings = f'Warnings: {", ".join(descr)}'
        else:
            warnings = None
        return '\n'.join(ret), warnings

    def format_alias_list(self, all_aliases, verbose=False):
        return '\n'.join([f'{k} -> {v}' for k, v in all_aliases.items()]), None

    def format_licenses(self, licenses, verbose=False):
        licenses.sort()
        return '\n'.join(licenses), None

    def format_operators(self, operators, verbose=False):
        return '\n'.join([f'{k} -> {v}' for k, v in operators.items()]), None

    def format_ambiguities(self, ambiguities, verbose=False):
        ret = []
        for k, v in ambiguities.items():
            ret += [f'{a} -> {k}' for a in v['aliases']]
        return '\n'.join(ret), None

    def format_compounds(self, compounds, verbose=False):
        ret = []
        for compound in compounds:
            ret += [f'{a} -> {compound["spdxid"]}' for a in compound['aliases']]
        return '\n'.join(ret), None

    def format_error(self, error, verbose=False):
        return f'Error: {error}', None

    def format_license_complete(self, lic, verbose=False):
        ret_str = []
        ret_str.append(f'{lic["name"]}')
        ret_str.append(f'    spdxid:           {lic["spdxid"]}')
        ret_str.append(f'    spdx_url:         https://spdx.org/licenses/{lic["spdxid"]}.html')
        ret_str.append(f'    scancode_key:     {lic["scancode_key"]}')
        ret_str.append(f'    scancode_db_url:  https://scancode-licensedb.aboutcode.org/{lic["scancode_key"]}.html')
        ret_str.append(f'    aliases:          {", ".join(lic["aliases"])}')
        if 'compatibility_as' in lic:
            ret_str.append(f'    compatibility_as: {lic["compatibility_as"]}')
        if 'comment_as' in lic:
            ret_str.append(f'    comment:          {lic["comment"]}')

        return "\n".join(ret_str), None
